from enum import IntEnum


class FacadeErrorCode(IntEnum):
    none = 0
    unknown = 1
    protocol = 2
    invalid_type = 3
    invalid_state = 4
    invalid_input = 5
    reader_not_ready = 6
    writer_not_ready = 7


class FacadeError(Exception):
    def __init__(self, invocation: str, code: FacadeErrorCode):
        self.invocation = invocation
        self.code = code

    def __str__(self) -> str:
        return f"{self.invocation} return with error {self.code}"



from cffi import FFI
import os
import sys

from .libfacade_h import facade_h

LIBFACADE_LIB_DIR = os.getenv("LIBFACADE_DIR")

if sys.platform == 'darwin':
    if LIBFACADE_LIB_DIR is None:
        LIBFACADE_LIB_DIR = '/Applications/Facade.app/Contents/Frameworks/libfacade.dylib.framework/'

os.environ['DYLD_LIBRARY_PATH'] = f"{LIBFACADE_LIB_DIR}" \
                                  f"{':' +  os.environ['DYLD_LIBRARY_PATH'] if 'DYLD_LIBRARY_PATH' in os.environ else ''}"

ffi = FFI()
ffi.cdef(facade_h)

libfacade = ffi.dlopen('facade')



facade_h = """
typedef enum {
    facade_device_type_video = 0,       
} facade_device_type;
typedef enum {
    facade_error_none = 0,             
    facade_error_unknown = 1,          
    facade_error_protocol = 2,         
    facade_error_invalid_type = 3,     
    facade_error_invalid_state = 4,    
    facade_error_invalid_input = 5,    
    facade_error_reader_not_ready = 6, 
    facade_error_writer_not_ready = 7, 
    facade_error_not_installed = 8,    
    facade_error_not_initialized = 9,  
    facade_error_not_found = 10,       
} facade_error_code;
typedef uint64_t facade_version;
typedef struct facade_device_info facade_device_info;
typedef struct facade_device_info {
    facade_device_info *next;          
    facade_device_type type;           
    const char *uid;                   
    const char *name;                  
    uint32_t width;                    
    uint32_t height;                   
    uint32_t frame_rate;               
} facade_device_info;
typedef struct {
    facade_version api_version;
    facade_device_info *devices;
} facade_state;
typedef struct facade_device_data facade_device_data;
typedef struct facade_device facade_device;
typedef struct facade_device {
    facade_device *next;               
    facade_device_type type;           
    const char *uid;                   
    char const *name;                  
    uint32_t width;                    
    uint32_t height;                   
    uint32_t frame_rate;               
    facade_device_data *data;          
} facade_device;
typedef void (*facade_callback)(void *context);
facade_error_code facade_init(void);
facade_error_code facade_read_state(facade_state **p);
facade_error_code facade_write_state(facade_state *p);
facade_error_code facade_on_state_changed(facade_callback callback, void *context);
facade_error_code facade_dispose_state(facade_state **p);
facade_error_code facade_list_devices(facade_device **p);
facade_error_code facade_find_device_by_uid(char const *uid, facade_device **p);
facade_error_code facade_find_device_by_name(char const *name, facade_device **p);
facade_error_code facade_dispose_device(facade_device **p);
facade_error_code facade_create_device(facade_device_info *options);
facade_error_code facade_edit_device(char const *uid, facade_device_info *options);
facade_error_code facade_delete_device(char const *uid);
facade_error_code facade_on_device_changed(facade_device *, facade_callback callback, void *context);
facade_error_code facade_read_open(facade_device *p);
facade_error_code facade_read_callback(facade_device *p, facade_callback callback, void *context);
facade_error_code facade_read_frame(facade_device *p, void **buffer, size_t *buffer_size);
facade_error_code facade_read_close(facade_device *p);
facade_error_code facade_write_open(facade_device *);
facade_error_code facade_write_callback(facade_device *p, facade_callback callback, void *context);
facade_error_code facade_write_frame(facade_device *p, void *buffer, size_t buffer_size);
facade_error_code facade_write_close(facade_device *p);
void free(void *);
"""



from abc import ABCMeta, abstractmethod
from typing import Union, Literal, Callable, List, Optional

from .facade_device_type import FacadeDeviceType
from .facade_error_code import FacadeError, FacadeErrorCode
from .libfacade import ffi, libfacade


class FacadeDevice(metaclass=ABCMeta):
    def __init__(self, pointer):
        self.__read_callback = None
        self.__write_callback = None
        self.__changed_callback = None
        self._pointer = pointer

    def __str__(self) -> str:
        return f"FacadeDevice{{uid={self.uid}}}"

    def __del__(self):
        double_pointer = ffi.new(cdecl='facade_device**', init=self._pointer)
        libfacade.facade_dispose_device(double_pointer)

    @property
    @abstractmethod
    def type(self) -> FacadeDeviceType:
        pass

    @property
    def uid(self) -> int:
        return self._pointer.uid

    def read_callback(self, value: Callable[[], None]) -> None:
        if value is None:
            self.__read_callback = None
            libfacade.facade_read_callback(self._pointer, ffi.NULL, ffi.NULL)
        else:
            def trap(_arg0):
                value()
            callback = ffi.callback(cdecl=ffi.typeof("void(*)(void *)"),
                                    python_callable=trap)
            self.__read_callback = callback
            libfacade.facade_read_callback(self._pointer, callback, ffi.NULL)

    def write_callback(self, value: Callable[[], None]) -> None:
        if value is None:
            self.__write_callback = None
            libfacade.facade_write_callback(self._pointer, ffi.NULL, ffi.NULL)
        else:
            def trap(_arg0):
                value()
            callback = ffi.callback(cdecl=ffi.typeof("void(*)(void *)"),
                                    python_callable=trap)
            self.__write_callback = callback
            libfacade.facade_write_callback(self._pointer, callback, ffi.NULL)

    def changed_callback(self, value: Callable[[], None]) -> None:
        if value is None:
            self.__changed_callback = None
            libfacade.facade_on_device_changed(self._pointer, ffi.NULL, ffi.NULL)
        else:
            def trap(_arg0):
                value()
            callback = ffi.callback(cdecl=ffi.typeof("void(*)(void *"),
                                    python_callable=trap)
            self.__changed_callback = callback
            libfacade.facade_on_device_changed(self._pointer, callback, ffi.NULL)

    def open(self, mode: Union[Literal['r'], Literal['w']]) -> None:
        if 'r' in mode:
            code = libfacade.facade_read_open(self._pointer)
            if code != FacadeErrorCode.none:
                raise FacadeError('facade_read_open', code)
        elif 'w' in mode:
            code = libfacade.facade_write_open(self._pointer)
            if code != FacadeErrorCode.none:
                raise FacadeError('facade_write_open', code)

    def close(self) -> None:
        libfacade.facade_read_close(self._pointer)
        libfacade.facade_write_close(self._pointer)

    @staticmethod
    def create(device) -> 'FacadeDevice':
        from .video_facade_device import VideoFacadeDevice
        return VideoFacadeDevice(device) if device.type == 0 else FacadeDevice(device)

    @staticmethod
    def by_uid(uid: int) -> Optional['FacadeDevice']:
        pointer = ffi.new(cdecl='facade_device**', init=ffi.NULL)
        code = libfacade.facade_find_device_by_uid(uid.encode(), pointer)

        if code != FacadeErrorCode.none:
            raise FacadeError(invocation="facade_find_device_by_uid", code=code)

        return FacadeDevice.create(pointer[0])

    @staticmethod
    def by_name(name: str) -> Optional['FacadeDevice']:
        pointer = ffi.new(cdecl='facade_device**', init=ffi.NULL)
        code = libfacade.facade_find_device_by_name(name.encode(), pointer)

        if code != FacadeErrorCode.none:
            raise FacadeError(invocation="facade_find_device_by_name", code=code)

        return FacadeDevice.create(pointer[0])

    @staticmethod
    def list() -> List['FacadeDevice']:
        list_head = ffi.new('facade_device **')
        libfacade.facade_list_devices(list_head)
        devices = []

        node = list_head[0]
        visited = False

        while node != ffi.NULL and (node != list_head[0] or not visited):
            devices.append(FacadeDevice.create(node))
            node = node.next
            visited = True

        return devices



from typing import Optional

from .facade_device import FacadeDevice
from .facade_device_type import FacadeDeviceType
from .facade_error_code import FacadeError, FacadeErrorCode
from .libfacade import ffi, libfacade


class VideoFacadeDevice(FacadeDevice):
    def __str__(self) -> str:
        return f"FacadeDevice{{uid={self.uid} width={self.width}, height={self.width}, frame_rate={self.frame_rate}}}"

    @property
    def type(self) -> FacadeDeviceType:
        return FacadeDeviceType.video

    @property
    def width(self) -> int:
        return self._pointer.width

    @property
    def height(self) -> int:
        return self._pointer.height

    @property
    def frame_rate(self) -> int:
        return self._pointer.frame_rate

    @property
    def frame_byte_size(self) -> int:
        return self.width * self.height * 4

    def read_frame(self, buffer: Optional[bytearray]) -> Optional[ffi.buffer]:
        if buffer is None:
            buffer_size = self.frame_byte_size
            native_buffer = ffi.new(f'uint8_t[{buffer_size}]')
        else:
            if len(buffer) < self.frame_byte_size:
                raise ValueError(f"buffer is too small (minimum size is {self.frame_byte_size})")
            buffer_size = len(buffer)
            native_buffer = ffi.from_buffer(cdecl='uint8_t *', python_buffer=buffer)

        code = libfacade.facade_read(self._pointer, native_buffer, buffer_size)

        if code not in (FacadeErrorCode.none, FacadeErrorCode.reader_not_ready):
            raise FacadeError('facade_read_frame', code)

        return ffi.buffer(native_buffer) if code == FacadeErrorCode.none else None

    def write_frame(self, buffer: Optional[bytearray]):
        buffer_size = self.width * self.height * 4
        if len(buffer) < buffer_size:
            raise ValueError("buffer is of wrong size")

        code = libfacade.facade_write_frame(self._pointer, ffi.from_buffer('void *', buffer), len(buffer))

        if code != FacadeErrorCode.none:
            raise FacadeError('facade_write_frame', code)



from setuptools import setup

setup(
    name='pyfacade',
    version='1.0.0rc1',
    description='A Python interface to the libktx library',
    author='Shukant Pal',
    author_email='facade@palcor.net',
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    include_package_data=True,
    install_requires=["cffi>=1.15.1"],
    packages=['facade'],
    package_dir={'facade': 'facade'},
    setup_requires=["cffi>=1.15.1"],
    url='https://source.shukantpal.com/Palco/Facade'
)
